baseVersion = '6.1.0'
cgcloudVersion = '1.6.0a1.dev393'
version = '6.1.0-3f9cba3766e52866ea80d0934498f8c8f3129c3f'
cacheTag = 'cache-local-py3.10'
mainCacheTag = 'cache-master-py3.10'
distVersion = '6.1.0'
exactPython = 'python3.10'
python = 'python3.10'
dockerTag = '6.1.0-3f9cba3766e52866ea80d0934498f8c8f3129c3f-py3.10'
currentCommit = '3f9cba3766e52866ea80d0934498f8c8f3129c3f'
dockerRegistry = 'quay.io/ucsc_cgl'
dockerName = 'toil'
dirty = False
cwltool_version = '3.1.20240112164112'
